/*
    SdpBrowser
    Copyright 2006 Martin Storsj

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA

    Martin Storsj
    martin@martin.st
*/

#import "AppController.h"
#import <IOBluetoothUI/objc/IOBluetoothDeviceSelectorController.h>
#import <IOBluetooth/objc/IOBluetoothSDPServiceRecord.h>
#import <IOBluetooth/objc/IOBluetoothSDPDataElement.h>
#import <IOBluetooth/objc/IOBluetoothSDPUUID.h>
#include <stdio.h>

NSString* objectToString(id object) {
	if ([object isKindOfClass: [NSArray class]]) {
		int i;
		NSArray* array = object;
		NSMutableString *out = [NSMutableString stringWithCapacity: 100];
		for (i = 0; i < [array count]; i++) {
			id obj = [array objectAtIndex: i];
			if ([out length] > 0)
				[out appendFormat: @", { %@ }", objectToString(obj)];
			else
				[out appendFormat: @"{ %@ }", objectToString(obj)];
		}
		return out;
	} else if ([object isKindOfClass: [NSString class]]) {
		return [NSString stringWithFormat: @"\"%@\"", object];
	} else if ([object isKindOfClass: [NSNumber class]]) {
//		return [NSString stringWithFormat: @"%@", object];
		return [NSString stringWithFormat: @"%#llx", [object longLongValue]];
	} else if ([object isKindOfClass: [NSData class]]) {
		NSData* data = object;
		const uint8_t* bytes = [data bytes];
		int i;
		NSMutableString* str = [NSMutableString stringWithCapacity: 100];
		[str appendString: @"<"];
		for (i = 0; i < [data length]; i++) {
			[str appendFormat: @"%02x", bytes[i]];
		}
		[str appendString: @">"];
		return str;
	} else if ([object isKindOfClass: [IOBluetoothSDPDataElement class]]) {
		IOBluetoothSDPDataElement* element = object;
		NSArray* arrayVal = [element getArrayValue];
		NSString* strVal = [element getStringValue];
		NSNumber* numberVal = [element getNumberValue];
		NSData* dataVal = [element getDataValue];
		if (arrayVal) {
			return objectToString(arrayVal);
		} else if (strVal) {
			return objectToString(strVal);
		} else if (numberVal) {
			return objectToString(numberVal);
		} else if (dataVal) {
			return objectToString(dataVal);
		}
		return @"unknown";
	} else {
		return [NSString stringWithFormat: @"%@", object];
	}
}

id findSubviewOfClass(NSView* view, id class) {
	if ([view isKindOfClass: class])
		return view;
	NSArray* array = [view subviews];
	int i;
	for (i = 0; i < [array count]; i++) {
		NSView* subview = [array objectAtIndex: i];
		id ret = findSubviewOfClass(subview, class);
		if (ret != nil)
			return ret;
	}
	return nil;
}

@implementation AppController

- (IBAction)selectDevice:(id)sender
{
	browse = YES;
	[self showSelectDialog];
}

- (void) showSelectDialog {
	if (device) {
		NSAlert *alert = [[NSAlert alloc] init];
		[alert setAlertStyle: NSInformationalAlertStyle];
		[alert setMessageText: @"Busy"];
		[alert setInformativeText: @"Another SDP query is already started"];
		[alert addButtonWithTitle: @"OK"];
//		[alert beginSheetModalForWindow: window modalDelegate: controller didEndSelector: @selector(alertDidEnd:returnCode:contextInfo:) contextInfo: NULL];
		[alert runModal];
		[alert release];
		return;
	}

	IOBluetoothDeviceSelectorController *controller = [[IOBluetoothDeviceSelectorController alloc] init];
//	[controller setOptions: kSearchOptionsAlwaysStartInquiry];
//	[controller setOptions: kSearchOptionsDiscardCachedResults];
	[controller setPrompt: @"Browse"];
	[controller setTitle: @"Browse device"];
	if (browse)
		[controller setDescriptionText: @"Select a bluetooth device to browse for services."];
	else
		[controller setDescriptionText: @"Select a bluetooth device which will be scanned for PC Suite services."];
//	[[controller window] setShowsResizeIndicator: NO];
	if ([controller runModal] == kIOBluetoothUISuccess) {
		device = [[controller getResults] objectAtIndex: 0];
		[device retain];
		int err;
		if ((err = [device performSDPQuery: self]) != kIOReturnSuccess) {
			[self showError: err];
			[device release];
			device = nil;
		}
	}
	[controller release];
	
}

- (IBAction)selectPcSuiteDevice:(id)sender
{
	browse = NO;
	[self showSelectDialog];
}

- (void)sdpQueryComplete:(IOBluetoothDevice *)dev status:(IOReturn)status
{
	if (status == kIOReturnSuccess) {
		if (browse) {
			[self showServiceWindow];
		} else {
			[self showPcSuiteWindow];
		}
	} else {
		[self showError: status];
	}
	[device release];
	device = nil;
}

- (void) showServiceWindow {
	NSArray* services = [device getServices];
	NSMutableString* string = [NSMutableString stringWithCapacity: 1000];
	if (services) {
		int i;
		for (i = 0; i < [services count]; i++) {
			IOBluetoothSDPServiceRecord *record = [services objectAtIndex: i];
			NSDictionary* attrs = [record getAttributes];
			NSArray* keys = [attrs allKeys];
			int j;
			for (j = 0; j < [keys count]; j++) {
				NSObject* key = [keys objectAtIndex: j];
				IOBluetoothSDPDataElement* val = [attrs objectForKey: key];
				NSString* str = [NSString stringWithFormat: @"%@: %@", key, objectToString(val)];
				[string appendFormat: @"%@\n", str];
			}
			[string appendString: @"\n"];
		}
	}

	NSNib* nib = [[NSNib alloc] initWithNibNamed: @"DeviceWindow" bundle: [NSBundle mainBundle]];
	NSArray* objects;
	[nib instantiateNibWithOwner: self topLevelObjects: &objects];
	int i;
	NSWindow* window = nil;
	for (i = 0; i < [objects count]; i++) {
		id obj = [objects objectAtIndex: i];
		if ([obj isKindOfClass: [NSWindow class]]) {
			window = obj;
		}
	}
	[nib release];

	NSTextView* textView = findSubviewOfClass([window contentView], [NSTextView class]);
	[textView setString: string];
	[window setTitle: [device getNameOrAddress]];
	[window makeKeyAndOrderFront: nil];
}

- (void) showPcSuiteWindow {
	IOBluetoothSDPUUID* ftpUUID = [IOBluetoothSDPUUID uuid16: 0x1106];
	const static uint8_t pcsuiteArray[] = {
		0x00, 0x00, 0x50, 0x05, 0x00, 0x00, 0x10, 0x00, 0x80, 0x00, 0x00, 0x02, 0xee, 0x00, 0x00, 0x01
	};
	IOBluetoothSDPUUID* pcsuiteUUID = [IOBluetoothSDPUUID uuidWithBytes: pcsuiteArray length: 16];
	IOBluetoothSDPServiceRecord* ftp = [device getServiceRecordForUUID: ftpUUID];
	IOBluetoothSDPServiceRecord* pcsuite = [device getServiceRecordForUUID: pcsuiteUUID];

	NSNib* nib = [[NSNib alloc] initWithNibNamed: @"PCSuiteInfo" bundle: [NSBundle mainBundle]];
	NSArray* objects;
	[nib instantiateNibWithOwner: self topLevelObjects: &objects];
	int i;
	NSWindow* window = nil;
	for (i = 0; i < [objects count]; i++)
		if ([[objects objectAtIndex: i] isKindOfClass: [NSWindow class]])
			window = [objects objectAtIndex: i];
	[nib release];
	
	NSTextField* handleField = [[window contentView] viewWithTag: 1];
	NSTextField* channelField = [[window contentView] viewWithTag: 2];
	BluetoothSDPServiceRecordHandle handle;
	if (ftp && [ftp getServiceRecordHandle: &handle] == kIOReturnSuccess) {
		[handleField setStringValue: [NSString stringWithFormat: @"%#x (%d)", handle, handle]];
	}
	BluetoothRFCOMMChannelID channel;
	if (pcsuite && [pcsuite getRFCOMMChannelID: &channel] == kIOReturnSuccess) {
		[channelField setStringValue: [NSString stringWithFormat: @"%d", channel]];
	}
	[window setTitle: [device getNameOrAddress]];
	[window makeKeyAndOrderFront: nil];

}

- (void) showError: (int) err {
	NSAlert *alert = [[NSAlert alloc] init];
	[alert setAlertStyle: NSInformationalAlertStyle];
	[alert setMessageText: @"SDP query error"];
	NSString *str = [NSString stringWithFormat: @"Bluetooth error %d", err];
	[alert setInformativeText: str];
	[alert addButtonWithTitle: @"OK"];
//	[alert beginSheetModalForWindow: window modalDelegate: controller didEndSelector: @selector(alertDidEnd:returnCode:contextInfo:) contextInfo: NULL];
	[alert runModal];
	[alert release];
}

@end
